# Gridlock - a collection of grid-based strategy games
# Copyright (c) 2009 Brian Nenninger (brian@dozingcatsoftware.com)
# Gridlock comes with ABSOLUTELY NO WARRANTY, and is released under version 3 of the GNU GPL.

import random
from Game import GameState
import PGUtils

NUM_PIECES = 3**4 # count, color, shape, opacity

START_POSITIONS = [(i,j) for i in xrange(4) for j in xrange(1,4)]
EXTRA_POSITIONS = ((0,0), (1,0), (2,0))
NOMATCH_POSITION = (3,0)
START_WALL_POSITIONS = ((0,0), (1,0), (2,0), (3,0))

def int_to_vector(x):
    return (x//27, (x//9)%3, (x//3)%3, x%3)
    
def check_set(numbers):
    """ Returns True if the numbers represent vectors with all unique or all different values in each position """
    vectors = [int_to_vector(n) for n in numbers]
    veclen = len(vectors)
    # for each index, get number of unique values in the vectors
    for i in xrange(len(vectors[0])):
        pvals = len(set(v[i] for v in vectors))
        if pvals!=1 and pvals!=veclen: 
            return False
    return True
        

class SetsPGGameState(GameState):
    def __init__(self, dimensions=None, params=None, players=2, reset=True):
        GameState.__init__(self, dimensions, params, 1, reset)

    def reset(self):
        GameState.reset(self)
        # fill the first 12 cells
        self.extra_state = {'pieces_used': len(START_POSITIONS)}
        # shuffle the (count, color, shape, opacity) values
        self.extra_state['all_pieces'] = random.sample(xrange(NUM_PIECES), NUM_PIECES)
        for index, p in enumerate(START_POSITIONS):
            self.grid[p] = self.extra_state['all_pieces'][index]
        for p in START_WALL_POSITIONS:
            self.grid[p] = -1
            
    def player_score(self, pnum):
        if pnum!=1: return 0
        return self.extra_state['pieces_used'] - sum(1 for pos,val in self.enumerate_all() if val>=0)
        
    def successors(self):
        found_set = False
        # check all triples of non-walls
        positions = [pos for pos,val in self.enumerate_all() if val>=0]
        for p1 in positions:
            for p2 in positions:
                for p3 in positions:
                    if p1!=p2 and p2!=p3 and p3!=p1:
                        move = (p1, p2, p3)
                        nextgrid = self.grid.copy()
                        # changes to nextgrid affect nextstate
                        nextstate = self.next_state_with_grid(nextgrid)
                        if check_set((self.grid[p1], self.grid[p2], self.grid[p3])):
                            found_set = True
                            # if "extra" positions are present, fill in the main cells
                            if self.grid[EXTRA_POSITIONS[0]]>=0:
                                for main_pos, extra_pos in zip( (p for p in move if p in START_POSITIONS),
                                                                (p for p in EXTRA_POSITIONS if p not in move)):
                                    nextgrid[main_pos] = nextgrid[extra_pos]
                                for pos in EXTRA_POSITIONS: 
                                    nextgrid[pos] = -1
                            else:
                                if self.extra_state['pieces_used']<len(self.extra_state['all_pieces']):
                                    # add new pieces from "deck"
                                    for index, pos in enumerate(move):
                                        nextgrid[pos] = self.extra_state['all_pieces'][self.extra_state['pieces_used']+index]
                                    nextstate.extra_state['pieces_used'] += len(move)
                                else:
                                    # no pieces left, fill in walls
                                    for p in move: 
                                        nextgrid[p] = -1
                        # if no match, no change to grid
                        yield (move, nextstate)
        # if pieces remaining, allow click on corner wall
        if self.extra_state['pieces_used']<len(self.extra_state['all_pieces']):
            # show extra pieces if no existing set (replace found_set with 0 to always show)
            if not found_set:
                nextgrid = self.grid.copy()
                for index,pos in enumerate(EXTRA_POSITIONS):
                    nextgrid[pos] = self.extra_state['all_pieces'][self.extra_state['pieces_used']+index]
                nextstate = self.next_state_with_grid(nextgrid)
                nextstate.extra_state['pieces_used'] += len(EXTRA_POSITIONS)
                yield ((NOMATCH_POSITION,), nextstate)
            else:
                # set exists; allow click but don't change anything
                yield ((NOMATCH_POSITION,), self.next_state_with_grid(self.grid.copy()))

    def is_game_over(self):
        # game is over if no pieces left in the "deck", and no valid sets on the board
        if self.extra_state['pieces_used']<len(self.extra_state['all_pieces']): 
            return False
        return not any(check_set(self.grid[pos] for pos in move) for move,state in self.successors())
    
    def animations_for_move(self, move, newstate):                
        import PGPieceAnimation
        animations = []
        if len(move)>1 and check_set(self.grid[pos] for pos in move):
            animations.extend(PGPieceAnimation.remove_animation(pos) for pos in move)
            # if extra positions were filled, animate moving to fill in base positions
            if self.grid[EXTRA_POSITIONS[0]]>=0:
                for main_pos, extra_pos in zip( (p for p in move if p in START_POSITIONS),
                                                (p for p in EXTRA_POSITIONS if p not in move)):
                    animations.append(PGPieceAnimation.move_animation(extra_pos, main_pos))
        return PGPieceAnimation.AnimationGroup(animations)

from GenericAI import GenericAI

class SetsPGAI(GenericAI):
    def game_utility(self, game):
        # Add 1 point for using extra squares so they'll be used if no match.
        # Utility has to be negated because GenericAI.alphabeta assumes it's trying 
        # to minimize the "next" player's utility.
        utility = game.player_score(1)
        if game.grid[EXTRA_POSITIONS[0]]>=0:
            utility += 1
        return -utility

